class Color < Gosu::Color
  def initialize(red, green, blue, alpha = 255)
    super(alpha.to_i, red.to_i, green.to_i, blue.to_i)
  end
  def set(red, green, blue, alpha = 255)
    self.red, self.green, self.blue, self.alpha = red, green, blue, alpha
  end
  def blend(color)
    self.clone.blend!(color)
  end
  def blend!(color)
    self.red  =(self.red  *(255-color.alpha)/255)+(color.red  *color.alpha/255)
    self.green=(self.green*(255-color.alpha)/255)+(color.green*color.alpha/255)
    self.blue =(self.blue *(255-color.alpha)/255)+(color.blue *color.alpha/255)
    self.alpha += (color.alpha * (255 - self.alpha) / 255)
    self
  end
  def to_a
    [red / 255.0, green / 255.0, blue / 255.0, alpha / 255.0]
  end
  def dup
    self.class.new(self.red, self.green, self.blue, self.alpha)
  end
  def clone
    self.class.new(self.red, self.green, self.blue, self.alpha)
  end
  # Dump to String (compatible with RGSS Color dump format)
  def _dump(marshal_depth = -1)
    [self.red, self.green, self.blue, self.alpha].pack('E4')
  end
  # Load from String (compatible with RGSS Color dump format)
  def self._load(data)
    new(*data.unpack('E4'))
  end
  White    = new(255, 255, 255)
  Shadow   = new(0, 0, 0, 160)
  Clear    = new(0, 0, 0, 0)
  Normal   = White
  Disabled = new(255, 255, 255, 128)
  System   = new(192, 224, 255)
  Crisis   = new(255, 255, 64)
  Knockout = new(255, 64, 0)
end

class Font
  @@default_name = 'Arial'
  @@default_size = 24
  @@default_bold = false
  @@default_italic = false
  @@default_shadow = false
  @@default_color = Color::Normal
  cattr_accessor :default_name, :default_size, :default_bold, :default_italic
  cattr_accessor :default_shadow, :default_color
  attr_accessor :name, :size, :bold, :italic, :shadow, :color
  def initialize(name = @@default_name, size = @@default_size)
    @name = name
    @size = size
    @bold = @@default_bold
    @italic = @@default_italic
    @shadow = @@default_shadow
    @color = @@default_color.dup
    recache
  end
  def name=(name)
    @name = name
    recache
    name
  end
  def size=(size)
    @size = size
    recache
    size
  end
  def recache
    @gosufont = Gosu::Font.new($GameWindow, @name, @size)
  end
  def draw(text, x, y, z = 0, factor_x = 1.0, factor_y = 1.0)
    text = text.to_s.gsub('<', '&lt;').gsub('>', '&gt;') # Literalize tags
    text = "<b>#{text}</b>" if @bold
    text = "<i>#{text}</i>" if @italic
    if @shadow
      sdist = height / 10
      c = Color::Shadow
      @gosufont.draw(text, x + sdist, y + sdist, z, factor_x, factor_y, c)
    end
    @gosufont.draw(text, x, y, z, factor_x, factor_y, @color)
  end
end

class Gosu::Image
  # @note This function does not clip the drawn image within the provided
  #       width and height; the caller should use Gosu::Window#clip_to
  def draw_tiled(x, y, w, h, z, fx=1.0, fy=1.0, c=0xffffffff, mode=:default)
    # Max width and height
    max_width  = x + w
    max_height = y + h
    # Amount to change X and Y by
    dx = self.width * fx
    dy = self.height * fy
    
    draw_x = x
    while draw_x < max_width
      draw_y = y
      while draw_y < max_height
        draw(draw_x.to_i, draw_y.to_i, z, fx, fy, c, mode)
        draw_y += dy
      end
      draw_x += dx
    end
  end
  # From now on :trect refers to "Texplay's rect function". This is provided so
  # that it is always available by the same name whether or not the image is of
  # the Bitmap class (which has a :rect method with different meaning)
  alias trect :rect
  # Reimplement Texplay.clear to use :trect
  def clear(options = {})
    options = {
      :color => :alpha,
      :fill => true
    }.merge!(options)

    capture {
      trect 0, 0, width - 1, height - 1, options
      
      self
    }
  end
end

class ImageLoader
  attr_accessor :rows, :columns, :to_blob
  def initialize(filename)
    @rows, @columns = 5, 5
    @to_blob = "\0" * 4 * @rows * @columns
  end
end

class Bitmap < Gosu::Image
  include Disposable # :dispose does nothing, but must exist; disposed with GC
  attr_accessor :font
  def initialize(filename, height = nil)
    if height.nil?
      for path in Settings::SEARCH_PATHS
        unless File.file?(fname = File.join(path, filename))
          possible = Dir.glob(fname + '*')
          if possible.length > 0
            filename = possible[0]
            break
          end
        end
      end
      unless ['.png', ',bmp'].include?(File.extname(filename))
        filename = ImageLoader.new(filename)
      end
      super($GameWindow, filename)
    else
      super($GameWindow, TexPlay::EmptyImageStub.new(filename, height))
    end
    @font = Font.new
  end
  def dup
    output = self.class.allocate
    Gosu::Image.instance_method(:initialize).bind(output).call($GameWindow,self)
    output
  end
  def clone
    output = self.dup
    output.freeze if self.frozen?
    output.taint if self.tainted?
    output
  end
  def rect
    Rect.new(0, 0, width, height)
  end
  def blt(x, y, src_bitmap, src_rect = nil, opacity = 255)
    src_rect = src_bitmap.rect if src_rect.nil?
    opacity /= 255.0
    splice(src_bitmap, x, y, :alpha_blend => true, :crop => src_rect.to_a(true),
           :color_control => lambda do |c1, c2|
      c2[3] *= opacity
      c2
    end)
  end
  def stretch_blt(dest_rect, src_bitmap, src_rect, opacity = 255)
    src_rect = src_bitmap.rect if src_rect.nil?
    opacity /= 255.0
    fixx, fixy = dest_rect.x + src_rect.x, dest_rect.y + src_rect.y
    factor_x = dest_rect.width / src_rect.width
    factor_y = dest_Rect.height / src_rect.height
    trect(dest_rect.x, dest_rect.y, dest_rect.x + dest_rect.width - 1,
          dest_rect.y + dest_rect.height - 1, :fill => true,
          :alpha_blend => true, :color_control => lambda do |c, x2, y2|
              src_x, src_y = (x2 - fixx) / factor_x, (y2 - fixy) / factor_y
              color = src_bitmap.get_pixel(src_x, src_y).to_a
              color[3] *= opacity
              color
    end)
  end
  def fill_rect(x, y, width = nil, height = nil, color = nil)
    x, y, width, height, color = *(x.to_a << y) if width.nil?
    trect(x, y, x + width, y + height, :color => color.to_a, :fill => true)
  rescue NoMethodError
    errstr = 'Arguments must be x, y, width, height, color OR rect, color'
    raise ArgumentError, errstr
  end
  def gradient_fill_rect(x, y, width=nil, height=nil, c1=nil, c2=nil, vert=nil)
    x, y, width, height, c1, c2, vert = *(x.to_a + [y, width, height]) if c1.nil?
    c1, c2 = c1.to_a, c2.to_a
    control = (vert ? lambda do |c, x_, y_|
      d = (y_ - y) / height
      c[0] = c1[0] * (1.0 - d) + c2[0] * d
      c[1] = c1[1] * (1.0 - d) + c2[1] * d
      c[2] = c1[2] * (1.0 - d) + c2[2] * d
      c[3] = c1[3] * (1.0 - d) + c2[3] * d
      c
    end : lambda do |c, x_, y_|
      d = (x_ - x) / width
      c[0] = c1[0] * (1.0 - d) + c2[0] * d
      c[1] = c1[1] * (1.0 - d) + c2[1] * d
      c[2] = c1[2] * (1.0 - d) + c2[2] * d
      c[3] = c1[3] * (1.0 - d) + c2[3] * d
      c
    end)
    trect(x, y, x + width, y + height, :color_control => control, :fill => true)
  rescue NoMethodError
    errstr = 'Arguments must be x, y, width, height, c1, c2 OR rect, c1, c2'
    raise ArgumentError, errstr
  end
  def clear_rect(x, y = nil, width = nil, height = nil)
    x, y, width, height = *x.to_a if y.nil?
    trect(x, y, x + width, y + height, color => :alpha, :fill => true)
  rescue NoMethodError
    errstr = 'Arguments must be x, y, width, height'
    raise ArgumentError, errstr
  end
  def get_pixel(x, y)
    return nil if (ar = super(x, y)).nil?
    Color.new(ar[0] * 255, ar[1] * 255, ar[2] * 255, ar[3] * 255)
  end
  def set_pixel(x, y, color)
    super(x, y, :color => color.to_a)
  end
  def hue_change(hue)
    each do |c|
      color = Color.new(c[0] * 255, c[1] * 255, c[2] * 255, c[3] * 255)
      color.hue += hue
      c[0..3] = color.to_a
    end
  end
  def blur(radius = 2)
    each do  |c, x, y|
      num_pixels = 0
      total = [0, 0, 0, 0]
      circle x, y, radius, :color_control => lambda do |v|
        if v
          total[0] += v[0]
          total[1] += v[1]
          total[2] += v[2]
          total[3] += v[3]
          num_pixels += 1
        end
        :none
      end
      c[0] = total[0] / num_pixels
      c[1] = total[1] / num_pixels
      c[2] = total[2] / num_pixels
      c[3] = total[3] / num_pixels
    end
  end
  def radial_blur(angle, division)
  end
  def draw_text(x, y, width = 0, height = nil, str = nil, align = 0)
    x, y, width, height, str, align = *(x.to_a + [y, width]) if height.nil?
#    $GameWindow.render_to_image(self) do
#      @font.draw(str, x, y)
#    end
    str = str.to_s.gsub('<', '&lt;').gsub('>', '&gt;') # Literalize tags
    str = "<b>#{str}</b>" if @font.bold
    str = "<i>#{str}</i>" if @font.italic
    image = Gosu::Image.from_text($GameWindow, str, @font.name, @font.size)
    
    sdist = @font.size / 10
    x += (width - (@font.shadow ? image.width+sdist : image.width)) * align / 2
    y += (height - (@font.shadow ? image.height+sdist : image.height)) / 2
    
    if @font.shadow
      splice(image, x + sdist, y + sdist, :alpha_blend => true,
            :color_control => lambda do |c1, c2|
        c2[0] *= Color::Shadow.red   / 255.0
        c2[1] *= Color::Shadow.green / 255.0
        c2[2] *= Color::Shadow.blue  / 255.0
        c2[3] *= Color::Shadow.alpha / 255.0
        c2
      end)
    end
    
    splice(image,x,y,:alpha_blend => true, :color_control => lambda do |c1,c2|
      c2[0] *= @font.color.red   / 255.0
      c2[1] *= @font.color.green / 255.0
      c2[2] *= @font.color.blue  / 255.0
      c2[3] *= @font.color.alpha / 255.0
      c2
    end)
  rescue NoMethodError
    errstr = 'Arguments must be x, y, width, height, str[, align]'
    raise ArgumentError, errstr
  end
  def text_size(str)
    str = str.to_s.gsub('<', '&lt;').gsub('>', '&gt;') # Literalize tags
    str = "<b>#{str}</b>" if @font.bold
    image = Gosu::Image.from_text($GameWindow, str, @font.name, @font.size)
    Rect.new(0, 0, image.width, image.height)
  end
end
